/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "surface.pmh"

#define	FE_STA_DEBUG	FALSE
#define	FE_STA_PROFILE	FALSE

namespace fe
{

void SurfaceTrianglesAccessible::bind(
	sp<SurfaceAccessibleI> a_spSurfaceAccessibleI)
{
	m_spSurfaceAccessibleI=a_spSurfaceAccessibleI;
}

void SurfaceTrianglesAccessible::cache(void)
{
	//* populate surface with generic accessors

#if FE_STA_DEBUG
	feLog("SurfaceTrianglesAccessible::cache valid %d\n",
			m_spSurfaceAccessibleI.isValid());
#endif

#if FE_STA_PROFILE
	sp<Profiler> spProfiler(new Profiler("SurfaceTrianglesAccessible"));
	sp<Profiler::Profile> spProfilePrep(
			new Profiler::Profile(spProfiler,"Prep"));
	sp<Profiler::Profile> spProfileLoop(
			new Profiler::Profile(spProfiler,"Loop"));
	sp<Profiler::Profile> spProfileResolve(
			new Profiler::Profile(spProfiler,"Resolve"));
	spProfiler->begin();
	spProfilePrep->start();
#endif

	sp<SurfaceAccessorI> spElementUV=m_spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_vertex,SurfaceAccessibleI::e_uv,
			SurfaceAccessibleI::e_refuseMissing);
	const BWORD vertexUVs=spElementUV.isValid();
	if(!vertexUVs)
	{
#if FE_STA_DEBUG
		feLog("SurfaceTrianglesAccessible::cache no vertex uvs\n");
#endif
		spElementUV=m_spSurfaceAccessibleI->accessor(
				SurfaceAccessibleI::e_point,SurfaceAccessibleI::e_uv,
				SurfaceAccessibleI::e_refuseMissing);
#if FE_STA_DEBUG
		if(spElementUV.isNull())
		{
			feLog("SurfaceTrianglesAccessible::cache no point uvs\n");
		}
#endif
	}

	sp<SurfaceAccessorI> spVertices=m_spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_primitive,SurfaceAccessibleI::e_vertices,
			SurfaceAccessibleI::e_refuseMissing);

	sp<SurfaceAccessorI> spPointPosition=m_spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_point,SurfaceAccessibleI::e_position,
			SurfaceAccessibleI::e_refuseMissing);

	sp<SurfaceAccessorI> spPointNormal=m_spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_point,SurfaceAccessibleI::e_normal,
			SurfaceAccessibleI::e_refuseMissing);

	sp<SurfaceAccessorI> spPrimitiveNormal=m_spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_primitive,SurfaceAccessibleI::e_normal,
			SurfaceAccessibleI::e_refuseMissing);

	sp<SurfaceAccessorI> spPointColor=m_spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_point,SurfaceAccessibleI::e_color,
			SurfaceAccessibleI::e_refuseMissing);

	sp<SurfaceAccessorI> spPointAlpha=m_spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_point,"Alpha",
			SurfaceAccessibleI::e_refuseMissing);
	if(spPointAlpha.isNull())
	{
		spPointAlpha=m_spSurfaceAccessibleI->accessor(
				SurfaceAccessibleI::e_point,"alpha",
				SurfaceAccessibleI::e_refuseMissing);
	}

	sp<SurfaceAccessorI> spPrimitivePartition;
	if(!m_partitionAttr.empty())
	{
//		feLog("SurfaceTrianglesAccessible::cache partitionAttr \"%s\"\n",
//				m_partitionAttr.c_str());

		spPrimitivePartition=m_spSurfaceAccessibleI->accessor(
				SurfaceAccessibleI::e_primitive,m_partitionAttr.c_str(),
				SurfaceAccessibleI::e_refuseMissing);
	}

	const U32 primitiveCount=spVertices.isValid()? spVertices->count(): 0;

#if FE_STA_DEBUG
		feLog("SurfaceTrianglesAccessible::cache primitiveCount %d\n",
				primitiveCount);
#endif

	m_elements=primitiveCount;
	m_vertices=0;
	U32 elementAllocated=0;
	U32 vertexAllocated=0;

	U32 triangleIndex=0;
	U32 total=0;

	const SpatialVector zero(0.0,0.0,0.0);
	const Color white(1.0,1.0,1.0);

	setOptionalArrays(Arrays(e_arrayColor|e_arrayUV));

	//* pre-allocate estimate
	resizeFor(primitiveCount*2,primitiveCount*6,
			elementAllocated,vertexAllocated);

	const I32 pointCount=spPointPosition->count();

	m_hullPointMap.resize(pointCount);
	m_hullFacePointArray.resize(primitiveCount);

	for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
	{
		m_hullPointMap[pointIndex]= -1;
	}

#if FE_STA_PROFILE
	spProfileLoop->replace(spProfilePrep);
#endif

	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
		const U32 subCount=spVertices->subCount(primitiveIndex);

#if FE_STA_DEBUG
		feLog("SurfaceTrianglesAccessible::cache %d/%d tri %d/%d verts %d\n",
				primitiveIndex,primitiveCount,total,vertexAllocated,subCount);
#endif

		FEASSERT(subCount==3 || subCount==4);

		if(subCount<3)
		{
			continue;
		}

		if(subCount>4 && m_hullPointMap.size())
		{
			m_hullPointMap.clear();
			m_hullFacePointArray.clear();
		}

		if(primitiveIndex<m_hullFacePointArray.size() && subCount<4)
		{
			m_hullFacePointArray[primitiveIndex][3]= -1;
		}

		m_vertices+=(subCount==4)? 6: 3;

		if(m_vertices>vertexAllocated)
		{
			resizeFor(primitiveCount*2,m_vertices,
					elementAllocated,vertexAllocated);
		}

		FEASSERT(triangleIndex<elementAllocated);
		FEASSERT(total<vertexAllocated);

		I32 partitionIndex= -1;
		if(spPrimitivePartition.isValid())
		{
			const String partition=
					spPrimitivePartition->string(primitiveIndex);
			partitionIndex=lookupPartition(partition);

//			feLog("SurfaceTrianglesAccessible::cache partition %d \"%s\"\n",
//					partitionIndex,partition.c_str());
		}

		SpatialVector norm=spPrimitiveNormal.isValid()?
				spPrimitiveNormal->spatialVector(primitiveIndex): zero;

		m_pElementArray[triangleIndex]=Vector3i(total,3,FALSE);

		for(U32 subIndex=0;subIndex<subCount;subIndex++)
		{
			const I32 pointIndex=
					spVertices->integer(primitiveIndex,subIndex);

			if(pointIndex>=I32(vertexAllocated))
			{
				feLog("SurfaceTrianglesAccessible::cache"
						" corrupt point index\n");
				feLog("  primitiveIndex %d subIndex %d/%d pointIndex %d/%d\n",
						primitiveIndex,subIndex,subCount,
						pointIndex,vertexAllocated);
				continue;
			}

			FEASSERT(pointIndex<I32(vertexAllocated));

			if(pointIndex<I32(m_hullPointMap.size()))
			{
				FEASSERT(primitiveIndex<m_hullFacePointArray.size());

				m_hullFacePointArray[primitiveIndex][subIndex]=pointIndex;

				I32& rHullPointIndex=m_hullPointMap[pointIndex];
				if(rHullPointIndex<0)
				{
					rHullPointIndex=total;
				}
			}

			m_pVertexArray[total]=
					spPointPosition->spatialVector(pointIndex);
			m_pNormalArray[total]=(spPointNormal.isValid())?
					spPointNormal->spatialVector(pointIndex): norm;

			if(spElementUV.isNull())
			{
				m_pUVArray[total]=zero;
			}
			else if(vertexUVs)
			{
				m_pUVArray[total]=
						spElementUV->spatialVector(primitiveIndex,subIndex);
			}
			else
			{
				m_pUVArray[total]=spElementUV->spatialVector(pointIndex);
			}

			if(spPointColor.isNull())
			{
				m_pColorArray[total]=white;
			}
			else
			{
				m_pColorArray[total]=spPointColor->spatialVector(pointIndex);
			}

			if(spPointAlpha.isValid())
			{
				m_pColorArray[total][3]=spPointAlpha->real(pointIndex);
			}

			m_pTriangleIndexArray[total]=triangleIndex+(subIndex==3);
			m_pPointIndexArray[total]=pointIndex;
			m_pPrimitiveIndexArray[total]=primitiveIndex;
			m_pPartitionIndexArray[total]=partitionIndex;

#if FE_STA_DEBUG
			feLog("  %d/%d v %s n %s uv%s %s tri %d pt %d pr %d part %d\n",
					subIndex,subCount,
					c_print(m_pVertexArray[total]),
					c_print(m_pNormalArray[total]),
					spElementUV.isNull()? "": (vertexUVs? "(vtx)": "(pt)"),
					c_print(m_pUVArray[total]),
					m_pTriangleIndexArray[total],
					m_pPointIndexArray[total],
					m_pPrimitiveIndexArray[total],
					m_pPartitionIndexArray[total]);
#endif

			total++;

			if(subIndex==3)
			{
				triangleIndex++;

				FEASSERT(triangleIndex<elementAllocated);
				FEASSERT(total<vertexAllocated);

				m_pElementArray[triangleIndex]=
						Vector3i(total-1,3,FALSE);

				m_pVertexArray[total]=m_pVertexArray[total-4];
				m_pNormalArray[total]=m_pNormalArray[total-4];
				m_pUVArray[total]=m_pUVArray[total-4];
				m_pColorArray[total]=m_pColorArray[total-4];
				m_pTriangleIndexArray[total]=triangleIndex;
				m_pPointIndexArray[total]=m_pPointIndexArray[total-4];
				m_pPrimitiveIndexArray[total]=primitiveIndex;
				m_pPartitionIndexArray[total]=partitionIndex;
				total++;

				FEASSERT(total<vertexAllocated);

				m_pVertexArray[total]=m_pVertexArray[total-3];
				m_pNormalArray[total]=m_pNormalArray[total-3];
				m_pUVArray[total]=m_pUVArray[total-3];
				m_pColorArray[total]=m_pColorArray[total-3];
				m_pTriangleIndexArray[total]=triangleIndex;
				m_pPointIndexArray[total]=m_pPointIndexArray[total-3];
				m_pPrimitiveIndexArray[total]=primitiveIndex;
				m_pPartitionIndexArray[total]=partitionIndex;
				total++;
			}
		}

		triangleIndex++;
	}

#if FE_STA_PROFILE
	spProfileResolve->replace(spProfileLoop);
#endif

	m_elements=triangleIndex;
	m_vertices=total;

	if(elementAllocated!=triangleIndex || vertexAllocated!=m_vertices)
	{
		resizeArrays(triangleIndex,m_vertices);
	}

	if(spElementUV.isNull())
	{
		deallocate(m_pUVArray);
		m_pUVArray=NULL;
	}

	if(spPointColor.isNull() && spPointAlpha.isNull())
	{
		deallocate(m_pColorArray);
		m_pColorArray=NULL;
	}

	calcBoundingSphere();

#if FE_STA_PROFILE
	spProfileResolve->finish();
	spProfiler->end();

	feLog("%s:\n%s\n",spProfiler->name().c_str(),
			spProfiler->report().c_str());
#endif
}

sp<SurfaceI::GaugeI> SurfaceTrianglesAccessible::gauge(void)
{
	if(m_spSurfaceAccessibleI.isNull())
	{
		return sp<SurfaceI::GaugeI>(NULL);
	}

	sp<SurfaceAccessorI> spPointPosition=m_spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_point,SurfaceAccessibleI::e_position,
			SurfaceAccessibleI::e_refuseMissing);

	sp<SurfaceAccessorI> spVertices=m_spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_primitive,SurfaceAccessibleI::e_vertices,
			SurfaceAccessibleI::e_refuseMissing);

	if(spPointPosition.isNull() || spVertices.isNull())
	{
		return sp<SurfaceI::GaugeI>(NULL);
	}

	sp<Gauge> spGauge(new Gauge(sp<SurfaceI>(this)));

	if(m_spMesh.isNull())
	{
		const I32 pointCount=spPointPosition->count();
		const I32 primitiveCount=spVertices->count();

		Array<F64> points(pointCount*3);
		Array<U32> faces(primitiveCount*6);

		for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
		{
			const SpatialVector point=
					spPointPosition->spatialVector(pointIndex);

			const I32 pointIndex3=pointIndex*3;
			points[pointIndex3]=point[0];
			points[pointIndex3+1]=point[1];
			points[pointIndex3+2]=point[2];
		}

		I32 total=0;

		for(I32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
		{
			//* NOTE tris or quads

			const I32 subCount=spVertices->subCount(primitiveIndex);
			if(subCount<3 || subCount>4)
			{
				feLog("SurfaceTrianglesAccessible::gauge"
						" prim %d/%d subCount %d (not tri or quad)\n",
						primitiveIndex,primitiveCount,subCount);
				continue;
			}

			faces[total]=spVertices->integer(primitiveIndex,0);
			faces[total+1]=spVertices->integer(primitiveIndex,1);
			faces[total+2]=spVertices->integer(primitiveIndex,2);
			total+=3;

			if(subCount>3)
			{
				faces[total]=spVertices->integer(primitiveIndex,0);
				faces[total+1]=spVertices->integer(primitiveIndex,2);
				faces[total+2]=spVertices->integer(primitiveIndex,3);
				total+=3;
			}
		}

		faces.resize(total);

		m_spMesh=spGauge->createMesh(points,faces);
	}
	else
	{
		spGauge->setMesh(m_spMesh);
	}

	return spGauge;
}

sp<Geodesic::Mesh> SurfaceTrianglesAccessible::Gauge::createMesh(
	Array<F64>& a_rPoints,Array<U32>& a_rFaces)
{
	return m_geodesic.createMesh(a_rPoints,a_rFaces);
}

void SurfaceTrianglesAccessible::Gauge::setMesh(sp<Geodesic::Mesh> a_spMesh)
{
	m_geodesic.setMesh(a_spMesh);
}

sp<Geodesic::Mesh> SurfaceTrianglesAccessible::Gauge::mesh(void)
{
	return m_geodesic.mesh();
}

void SurfaceTrianglesAccessible::Gauge::pick(SpatialVector a_source,
	Real a_maxDistance)
{
	m_source=a_source;
	m_maxDistance=a_maxDistance;

	I32 faceIndex=0;

	sp<ImpactI> spImpact=m_spSurfaceI->nearestPoint(a_source,-1.0);
	if(spImpact.isValid())
	{
		//* TODO use lookup table to skip degenerate primitives
		faceIndex=spImpact->face();
	}

	m_geodesic.setSource(faceIndex,a_source);
	m_geodesic.setMaxDistance(a_maxDistance);
}

Real SurfaceTrianglesAccessible::Gauge::distanceTo(SpatialVector a_target)
{
	if(magnitudeSquared(a_target-m_source)>m_maxDistance*m_maxDistance)
	{
		return INFINITY;
	}

	I32 faceIndex=0;
	sp<ImpactI> spImpact=m_spSurfaceI->nearestPoint(a_target,-1.0);
	if(spImpact.isValid())
	{
		//* TODO use lookup table to skip degenerate primitives
		faceIndex=spImpact->face();
	}

	return m_geodesic.distanceTo(faceIndex,a_target);
}

} // namespace
