/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "usd.pmh"

#define FE_DHY_DEBUG			FALSE
#define FE_DHY_POINT_SHADOWS	TRUE

namespace fe
{

sp<DrawBufferI> DrawHydra::createBuffer(void)
{
	sp<DrawHydra::Buffer> spDrawBuffer(new DrawHydra::Buffer);

	return spDrawBuffer;
}

DrawHydra::Buffer::Buffer(void):
	m_pDriver(NULL),
	m_vao(0),
	m_multicolor(FALSE),
	m_firstColor(0,0,0,0)
{
}

DrawHydra::Buffer::~Buffer(void)
{
	releaseAll();
}

void DrawHydra::Buffer::releaseAll(void)
{
	if(m_vao)
	{
		glDeleteVertexArrays(1,&m_vao);
		m_vao=0;
	}

	if(m_pDriver)
	{
		delete m_pDriver;
		m_pDriver=NULL;
	}

	DrawOpenGL::Buffer::releaseAll();
}

DrawHydra::DrawHydra(void)
{
	m_alphaMin=0.1;				//* TODO param
	m_shadowResolution=8192;	//* TODO param
	m_shadowBias= -0.001;		//* TODO param
}

DrawHydra::~DrawHydra(void)
{
}

void DrawHydra::initialize(void)
{
#if FE_DHY_DEBUG
	feLog("DrawHydra::initialize\n");
#endif

	const char* shading=(const char *)glGetString(GL_SHADING_LANGUAGE_VERSION);
	const Real version=atof(shading);
	if(version<Real(4))
	{
		feLog("DrawHydra::initialize GLSL 4 required, found \"%s\"\n",shading);

		feX(e_cannotCreate,"DrawHydra::initialize","hardware not capable");
	}

#if FE_HYDRA_TEST_LIGHTING
	m_lightingShader.reset(new TestLightingShader());
#endif
}

void DrawHydra::flush(void)
{
#if FE_DHY_DEBUG
	feLog("DrawHydra::flush\n");
#endif
	DrawOpenGL::flush();
}

template <typename T>
static VtArray<T> _BuildArray(T values[], int numValues)
{
	VtArray<T> result(numValues);
	std::copy(values, values+numValues, result.begin());
	return result;
}

void DrawHydra::updateDriver(HydraDriver*& a_rpDriver)
{
	if(!a_rpDriver)
	{
		a_rpDriver = new HydraDriver(HdTokens->hull);

		//* NOTE otherwise, alpha below 0.5 just disappears
		a_rpDriver->GetRenderPassState()->SetAlphaToCoverageUseDefault(false);
//		a_rpDriver->GetRenderPassState()->SetAlphaToCoverageEnabled(false);
		a_rpDriver->GetRenderPassState()->SetAlphaThreshold(0.01);

		a_rpDriver->GetRenderPassState()->SetCullStyle(HdCullStyleNothing);

#if FE_HYDRA_TEST_LIGHTING
		a_rpDriver->GetRenderPassState()->SetLightingShader(
				m_lightingShader);
#else
		HydraDelegate& rDelegate=a_rpDriver->GetDelegate();
		rDelegate.SetRefineLevel(drawMode()->refinement());

		rDelegate.AddSimpleLightTask(SdfPath("/SimpleLightTask"));
		rDelegate.AddShadowTask(SdfPath("/ShadowTask"));
		rDelegate.AddRenderSetupTask(SdfPath("/RenderSetupTask"));
		rDelegate.AddRenderTask(SdfPath("/RenderTask"));
#endif
	}
}

void DrawHydra::updateLighting(HydraDriver* a_pDriver)
{
	sp<CameraDisplay> spCameraDisplay=view()->cameraDisplay();
	const Box2i viewport=spCameraDisplay->viewport();

	Matrix<4,4,Real> proj4x4;
	if(spCameraDisplay->projection()==CameraI::e_perspective)
	{
		Real fovy;
		Real nearplane;
		Real farplane;
		spCameraDisplay->getPerspective(fovy,nearplane,farplane);

		const GLfloat aspect=(GLfloat)width(viewport)/
				(GLfloat)height(viewport);
		proj4x4=perspective(fovy,Real(aspect),nearplane,farplane);
	}
	else
	{
		Real zoom;
		Vector2 center;
		spCameraDisplay->getOrtho(zoom,center);

		const Vector2 size(width(viewport),height(viewport));
		proj4x4=ViewCommon::computeOrtho(zoom,center,size);
	}

	const GfMatrix4d projMatrix(
			proj4x4[0],proj4x4[1],proj4x4[2],proj4x4[3],
			proj4x4[4],proj4x4[5],proj4x4[6],proj4x4[7],
			proj4x4[8],proj4x4[9],proj4x4[10],proj4x4[11],
			proj4x4[12],proj4x4[13],proj4x4[14],proj4x4[15]);

	Real view16[16];
	spCameraDisplay->cameraMatrix().copy16(view16);

	const GfMatrix4d viewMatrix(
			view16[0],view16[1],view16[2],view16[3],
			view16[4],view16[5],view16[6],view16[7],
			view16[8],view16[9],view16[10],view16[11],
			view16[12],view16[13],view16[14],view16[15]);

	a_pDriver->SetCamera(viewMatrix, projMatrix,
			GfVec4d(0, 0, width(viewport), height(viewport)));

	SpatialTransform cameraTransform;
	invert(cameraTransform,spCameraDisplay->cameraMatrix());

	const SpatialVector lightDirFromCamera(-1.0,-1.0,-3.0);
	SpatialVector lightDir;
	rotateVector(cameraTransform,lightDirFromCamera,lightDir);
	normalizeSafe(lightDir);

	m_brightness=fe::minimum(Real(1),fe::maximum(Real(0),m_brightness));
	m_contrast=fe::minimum(Real(2),fe::maximum(Real(0),m_contrast));

//	m_lightingShader->SetLightingStateFromOpenGL();

#if FE_DHD_TEST_LIGHTING
	m_lightingShader->SetCamera(viewMatrix, projMatrix);
	m_lightingShader->SetSceneAmbient(GfVec3f(m_brightness));
	m_lightingShader->SetLight(0,
			GfVec3f(lightDir[0],lightDir[1],lightDir[2]),GfVec3f(m_contrast));
	m_lightingShader->SetLight(1,GfVec3f(0,-1,0),GfVec3f(0));

	GlfSimpleLightingContextRefPtr const& rLightingContext =
			m_lightingShader->GetLightingContext();
	rLightingContext->SetCamera(viewMatrix, projMatrix);
	rLightingContext->SetUseLighting(true);
	rLightingContext->SetSceneAmbient(GfVec4f(0));

	rLightingContext->SetLights(lights);
#else
	GlfSimpleLightVector lights(1);
	GlfSimpleLight& rLight=lights[0];
	rLight.SetPosition(GfVec4f(-lightDir[0],-lightDir[1],-lightDir[2],0));
	rLight.SetAmbient(GfVec4f(m_brightness));
	rLight.SetDiffuse(GfVec4f(m_contrast,m_contrast,m_contrast,1));
	rLight.SetSpecular(GfVec4f(0.5));
	rLight.SetShadowResolution(m_shadowResolution);
	rLight.SetShadowBias(m_shadowBias);
//	rLight.SetShadowBlur(GfVec4f(0.1));

	rLight.SetHasShadow(drawMode()->shadows());

	const SdfPath lightPath("/Light0");

	HydraDelegate& rDelegate=a_pDriver->GetDelegate();
	if(!rDelegate.HasLight(lightPath))
	{
		rDelegate.AddLight(lightPath,rLight);
	}
	rDelegate.SetLight(lightPath,HdxLightTokens->params,
			VtValue(rLight));

	const SdfPath renderSetupPath("/RenderSetupTask");
	rDelegate.AdjustRenderSetupTask(renderSetupPath,drawMode()->lit());
#endif
}

void DrawHydra::drawPoints(const SpatialVector *vertex,
	const SpatialVector *normal,U32 vertices,
	BWORD multicolor,const Color *color,
	sp<DrawBufferI> spDrawBuffer)
{
//	if(vertices>8)
//	{
//		feLog("DrawHydra::drawPoints vertices %d buffer %p\n",
//				vertices,spDrawBuffer.raw());
//	}

	if(spDrawBuffer.isNull())
	{
//		if(vertices>8)
//		{
//			feLog("DrawHydra::drawPoints bypass %d\n",vertices);
//		}

		DrawOpenGL::drawPoints(vertex,normal,vertices,
				multicolor,color,spDrawBuffer);
		return;
	}

	sp<Buffer> spBuffer=spDrawBuffer;
	if(spBuffer.isNull())
	{
		feLog("DrawHydra::drawPoints incompatible buffer\n");
		return;
	}

	if(spBuffer->m_multicolor!=multicolor || spBuffer->m_firstColor!=color[0])
	{
		spBuffer->releaseAll();
	}

	HydraDriver*& rpDriver=spBuffer->m_pDriver;
	updateDriver(rpDriver);

	GLuint& rVao=spBuffer->m_vao;
	if(!rVao)
	{
		const I32 refinement=drawMode()->refinement();
		const BWORD refined=(refinement>0);

		HydraDelegate& rDelegate=rpDriver->GetDelegate();
		rDelegate.SetRefineLevel(refinement);

		spBuffer->m_multicolor=multicolor;
		spBuffer->m_firstColor=color[0];

		const I32 pointCount=vertices;

#if FE_DHY_DEBUG
		feLog("DrawHydra::drawPoints hull points %d refine %d\n",
				pointCount,refinement);
#endif

		GfVec3f* points=new GfVec3f[pointCount];

		//* NOTE Hydra does not currently seem to care about point normals
		//* (uses normals always facing camera)
		GfVec3f* normals=NULL;
		if(FALSE&& normal)
		{
			normals=new GfVec3f[pointCount];
		}

		HydraDelegate::Interpolation colorInterp=HydraDelegate::CONSTANT;
		VtValue vtColor;
		VtVec4fArray vtColorArray;
		if(multicolor)
		{
			colorInterp=HydraDelegate::VERTEX;
			vtColorArray.resize(pointCount);
		}
		else
		{
			const Color& rColor=color[0];

			vtColor=GfVec4f(rColor[0],rColor[1],rColor[2],
					fe::maximum(Real(rColor[3]),m_alphaMin));
		}

		SpatialVector min(FLT_MAX,FLT_MAX,FLT_MAX);
		SpatialVector max(FLT_MIN,FLT_MIN,FLT_MIN);

		for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
		{
			const SpatialVector& point=vertex[pointIndex];

			points[pointIndex].Set(point.temp());

			for(I32 m=0;m<3;m++)
			{
				if(min[m]>point[m])
				{
					min[m]=point[m];
				}
				if(max[m]<point[m])
				{
					max[m]=point[m];
				}
			}

			if(normals)
			{
				normals[pointIndex].Set(normal[pointIndex].temp());
			}

			if(multicolor)
			{
				const Color& rColor=color[pointIndex];
				Real alpha=rColor[3];

#if TRUE
				//* HACK boost alpha
				alpha=pow(alpha,0.3);
				if(alpha>0.01)
				{
					alpha=fe::maximum(m_alphaMin,fe::minimum(alpha,Real(1)));
				}
#endif

				vtColorArray[pointIndex]=
						GfVec4f(rColor[0],rColor[1],rColor[2],alpha);
			}
		}

		const SpatialVector size=max-min;
		const Real largest=fe::maximum(size[0],fe::maximum(size[1],size[2]));
		const Real defaultWidth=0.02*largest;	//* NOTE tweak

		VtVec3fArray normalArray;
		if(normals)
		{
			normalArray=_BuildArray(normals,pointCount);

			delete[] normals;
			normals=NULL;
		}

		if(multicolor)
		{
			vtColor = VtValue(vtColorArray);
		}

		const SdfPath id("/DrawHydra_points");

#if TRUE
		const VtValue vtWidth((float)defaultWidth);
		const HydraDelegate::Interpolation widthInterp=HydraDelegate::CONSTANT;
#else
		VtFloatArray vtWidthArray;
		vtWidthArray.resize(pointCount);
		for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
		{
			vtWidthArray[pointIndex]=(pointIndex%2)? 0.01f: 0.04f;
		}
		const VtValue vtWidth=VtValue(vtWidthArray);
		const HydraDelegate::Interpolation widthInterp=HydraDelegate::VERTEX;
#endif

		rDelegate.Remove(id);

		SdfPath instancerId;
		rDelegate.AddPoints(id,
				_BuildArray(points,pointCount),
				normalArray,
				vtColor,colorInterp,
				vtWidth,widthInterp,
				instancerId);

		//* NOTE point modes are all equivalent
		rDelegate.SetReprName(id, refined? HdTokens->refined: HdTokens->hull);

		//* NOTE newly generated VAO acquires current state when first bound
		glGenVertexArrays(1, &rVao);
		glBindVertexArray(rVao);
		glBindVertexArray(0);

#if FE_DHY_DEBUG
		feLog("DrawHydra::drawPoints new VAO %d\n",rVao);
#endif

		delete[] points;
		points=NULL;
	}

#if !FE_DHY_POINT_SHADOWS
	const BWORD shadowing=drawMode()->shadows();
	drawMode()->setShadows(FALSE);
#endif

	//* NOTE high resolution shadows don't seem to work well for point clouds
	m_shadowBias= -0.1;
	m_shadowResolution=256;

	updateLighting(rpDriver);

#if !FE_DHY_POINT_SHADOWS
	drawMode()->setShadows(shadowing);
#endif

	glBindVertexArray(rVao);

//	feLog("DrawHydra::drawPoints \"%s\"\n",c_print(drawStyle()));

	const Real alpha=color? color[0][3]: 1.0f;
	const BWORD lit=isLit();

	setup(drawStyle(),lit,twoSidedLighting(),FALSE,
			frontfaceCulling(),backfaceCulling(),zBuffering());

	const BWORD blending=(multicolor || alpha!=1.0f ||
			drawStyle()==DrawMode::e_ghost);
	if(blending && drawStyle()!=DrawMode::e_foreshadow)
	{
		pushBlending();
	}

	rpDriver->Draw();

	if(drawStyle()==DrawMode::e_foreshadow)
	{
		if(blending)
		{
			pushBlending();
		}

		setup(drawStyle(),lit,twoSidedLighting(),FALSE,
				frontfaceCulling(),backfaceCulling(),FALSE);

		rpDriver->Draw();

		setup(drawStyle(),lit,twoSidedLighting(),FALSE,
				frontfaceCulling(),backfaceCulling(),zBuffering());
	}

	if(blending)
	{
		glPopAttrib();
	}

	glBindVertexArray(0);
}

void DrawHydra::drawTriangles(const SpatialVector* vertex,
	const SpatialVector* normal,const Vector2f* texture,U32 vertices,
	StripMode strip,BWORD multicolor,const Color* color,
	const Array<I32>* hullPointMap,const Array<Vector4i>* hullFacePoint,
	sp<DrawBufferI> spDrawBuffer)
{
	if(drawStyle()==DrawMode::e_pointCloud || spDrawBuffer.isNull() ||
			!hullPointMap || !hullFacePoint)
	{
		DrawOpenGL::drawTriangles(vertex,normal,texture,vertices,
				strip,multicolor,color,
				hullPointMap,hullFacePoint,spDrawBuffer);
		return;
	}

	drawPolys(3,vertex,normal,texture,vertices,strip,
			multicolor,color,hullPointMap,hullFacePoint,spDrawBuffer);
}

void DrawHydra::drawLinesInternal(const SpatialVector* vertex,
	const SpatialVector* normal,const Vector2f* texture,U32 vertices,
	StripMode strip,BWORD multicolor,const Color* color,
	const Vector3i* element,U32 elementCount,
	sp<DrawBufferI> spDrawBuffer)
{
//	if(vertices>8)
//	{
//		feLog("DrawHydra::drawLinesInternal"
//				" vertices %d buffer %p elementCount %d\n  %s\n",
//				vertices,spDrawBuffer.raw(),elementCount,
//				c_print(drawMode()));
//	}

	//* HACK
//	if(spDrawBuffer.isNull())
//	{
//		spDrawBuffer=createBuffer();
//	}

	if(strip==e_strip && !element)
	{
		elementCount=1;
	}

	if(spDrawBuffer.isNull() || !elementCount)
	{
//		if(vertices>8)
//		{
//			feLog("DrawHydra::drawLinesInternal bypass %d\n",vertices);
//		}

		DrawOpenGL::drawLinesInternal(vertex,normal,texture,vertices,strip,
				multicolor,color,element,elementCount,spDrawBuffer);
		return;
	}

	sp<Buffer> spBuffer=spDrawBuffer;
	if(spBuffer.isNull())
	{
		feLog("DrawHydra::drawLinesInternal incompatible buffer\n");
		return;
	}

	if(spBuffer->m_multicolor!=multicolor || spBuffer->m_firstColor!=color[0])
	{
		spBuffer->releaseAll();
	}

	HydraDriver*& rpDriver=spBuffer->m_pDriver;
	updateDriver(rpDriver);

	GLuint& rVao=spBuffer->m_vao;
	if(!rVao)
	{
		const BWORD uvSpace=isUvSpace();
		if(uvSpace && !texture)
		{
			feLog("DrawHydra::drawLinesInternal"
					" vertices %d in UV space without UV coordinates\n",
					vertices);
			return;
		}

		HydraDelegate& rDelegate=rpDriver->GetDelegate();
		rDelegate.SetRefineLevel(drawMode()->refinement());

		spBuffer->m_multicolor=multicolor;
		spBuffer->m_firstColor=color[0];

//		const GfVec3f center=rDelegate.PopulateBasicTestSet();

		const I32 pointCount=vertices+2*elementCount;

#if FE_DHY_DEBUG
		feLog("DrawHydra::drawLinesInternal"
				" hull points %d faces %d refine %d\n",
				pointCount,elementCount,drawMode()->refinement());
#endif

		GfVec3f* points=new GfVec3f[pointCount];
		int* numVerts=new int[elementCount];
		I32 pointIndex=0;

		GfVec3f* normals=NULL;
		if(FALSE&& normal)
		{
			normals=new GfVec3f[pointCount];
		}

		HydraDelegate::Interpolation colorInterp=HydraDelegate::CONSTANT;
		VtValue vtColor;
		VtVec4fArray vtColorArray;
		if(multicolor)
		{
			colorInterp=HydraDelegate::VERTEX;
			vtColorArray.resize(pointCount);
		}
		else
		{
			const Color& rColor=color[0];

			vtColor=GfVec4f(rColor[0],rColor[1],rColor[2],
					fe::maximum(Real(rColor[3]),m_alphaMin));
		}

		SpatialVector min(FLT_MAX,FLT_MAX,FLT_MAX);
		SpatialVector max(FLT_MIN,FLT_MIN,FLT_MIN);

		for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
		{
			I32 startIndex=0;
			I32 subCount=vertices;

			if(element)
			{
				const Vector3i& rElement=element[elementIndex];
				startIndex=rElement[0];
				subCount=rElement[1];
			}

			const I32 cvCount=subCount+2;

			for(I32 subIndex=0;subIndex<subCount;subIndex++)
			{
				SpatialVector point;
				if(uvSpace)
				{
					const Vector2f& rUV=texture[startIndex+subIndex];
					set(point,rUV[0],rUV[1],Real(0));
				}
				else
				{
					point=vertex[startIndex+subIndex];
				}

				points[pointIndex+1+subIndex].Set(point.temp());

				for(I32 m=0;m<3;m++)
				{
					if(min[m]>point[m])
					{
						min[m]=point[m];
					}
					if(max[m]<point[m])
					{
						max[m]=point[m];
					}
				}
			}
			points[pointIndex]=points[pointIndex+1];
			points[pointIndex+cvCount-1]=points[pointIndex+cvCount-2];

			if(normals)
			{
				for(I32 subIndex=0;subIndex<subCount;subIndex++)
				{
					//* NOTE normals appear to be in eye space
//					normals[pointIndex+1+subIndex].Set(
//							normal[startIndex+subIndex].temp());
					normals[pointIndex+1+subIndex]=
							GfVec3f(0.0f,0.0f,1.0f);
				}
				normals[pointIndex]=normals[pointIndex+1];
				normals[pointIndex+cvCount-1]=
						normals[pointIndex+cvCount-2];
			}

			if(multicolor)
			{
				for(I32 subIndex=0;subIndex<subCount;subIndex++)
				{
					const Color& rColor=color[startIndex+subIndex];

					vtColorArray[pointIndex+1+subIndex]=GfVec4f(
							rColor[0],rColor[1],rColor[2],
							fe::maximum(Real(rColor[3]),m_alphaMin));
				}
				vtColorArray[pointIndex]=vtColorArray[pointIndex+1];
				vtColorArray[pointIndex+cvCount-1]=
						vtColorArray[pointIndex+cvCount-2];
			}

			numVerts[elementIndex]=cvCount;
			pointIndex+=cvCount;
		}

		const SpatialVector size=max-min;
		const Real largest=fe::maximum(size[0],fe::maximum(size[1],size[2]));
		const Real defaultWidth=1e-4*largest;	//* NOTE tweak

		FEASSERT(pointIndex==pointCount);

		VtVec3fArray normalArray;
		if(normals)
		{
			normalArray=_BuildArray(normals,pointCount);

			delete[] normals;
			normals=NULL;
		}

		if(multicolor)
		{
			vtColor = VtValue(vtColorArray);
		}

		const SdfPath id("/DrawHydra_curves");
		const TfToken basis=HdTokens->catmullRom;

#if TRUE
		const VtValue vtWidth((float)defaultWidth);
		const HydraDelegate::Interpolation widthInterp=HydraDelegate::CONSTANT;
#else
		VtFloatArray vtWidthArray;
		vtWidthArray.resize(pointCount);
		for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
		{
			vtWidthArray[pointIndex]=(pointIndex%2)? 0.01f: 0.04f;
		}
		const VtValue vtWidth= VtValue(vtWidthArray);
		const HydraDelegate::Interpolation widthInterp=HydraDelegate::VERTEX;
#endif

		rDelegate.Remove(id);

		SdfPath instancerId;
		rDelegate.AddBasisCurves(id,
				_BuildArray(points,pointCount),
				_BuildArray(numVerts,elementCount),
				normalArray,
				basis,
				vtColor,colorInterp,
				vtWidth,widthInterp,
				instancerId);

		const BWORD refined=(drawMode()->refinement()>0);

		//* NOTE choices:
		//*		hull (same as smoothHull, wire, wireOnSurf, refinedWire)
		//*		refined (same as refinedWireOnSurf)
		//* see: pxr/imaging/lib/hd/renderIndex.cpp (all same for points)
		rDelegate.SetReprName(id, refined? HdTokens->refined: HdTokens->hull);

		//* NOTE newly generated VAO acquires current state when first bound
		glGenVertexArrays(1, &rVao);
		glBindVertexArray(rVao);
		glBindVertexArray(0);

#if FE_DHY_DEBUG
		feLog("DrawHydra::drawLinesInternal new VAO %d\n",rVao);
#endif

		delete[] numVerts;
		numVerts=NULL;

		delete[] points;
		points=NULL;
	}

	updateLighting(rpDriver);

	glBindVertexArray(rVao);

//	feLog("DrawHydra::drawLinesInternal \"%s\"\n",c_print(drawStyle()));

	const BWORD pushLineWidth=(lineWidth()!=1.0f);
	const Real alpha=color? color[0][3]: 1.0f;
	const BWORD lit=isLit();

	setup(drawStyle(),lit,twoSidedLighting(),FALSE,
			frontfaceCulling(),backfaceCulling(),zBuffering());

	const BWORD blending=(multicolor || alpha!=1.0f ||
			drawStyle()==DrawMode::e_ghost);
	if(blending && drawStyle()!=DrawMode::e_foreshadow)
	{
		pushBlending();
	}

	rpDriver->Draw();

	if(drawStyle()==DrawMode::e_foreshadow)
	{
		if(pushLineWidth)
		{
			glPushAttrib(GL_LINE_BIT);
			glLineWidth(lineWidth());
		}

		if(blending)
		{
			pushBlending();
		}

		setup(drawStyle(),lit,twoSidedLighting(),FALSE,
				frontfaceCulling(),backfaceCulling(),FALSE);

		rpDriver->Draw();

		setup(drawStyle(),lit,twoSidedLighting(),FALSE,
				frontfaceCulling(),backfaceCulling(),zBuffering());

		if(pushLineWidth)
		{
			//* restore line width
			glPopAttrib();
		}
	}

	if(blending)
	{
		glPopAttrib();
	}

	glBindVertexArray(0);
}

void DrawHydra::drawPolys(U32 grain,const SpatialVector *vertex,
	const SpatialVector *normal,const Vector2f *texture,
	U32 vertices,StripMode strip,BWORD multicolor,const Color *color,
	const Array<I32>* hullPointMap,const Array<Vector4i>* hullFacePoint,
	sp<DrawBufferI> spDrawBuffer)
{
	const BWORD hasHull=(hullPointMap && hullFacePoint);

//	if(vertices>8)
//	{
//		feLog("DrawHydra::drawPolys"
//				" vertices %d grain %d hull %d buffer %p\n  %s\n",
//				vertices,grain,hasHull,spDrawBuffer.raw(),c_print(drawMode()));
//	}

	//* HACK
	if(spDrawBuffer.isNull())
	{
		spDrawBuffer=createBuffer();
	}

	if(grain<3 || grain>4 || spDrawBuffer.isNull())
	{
//		if(vertices>8)
//		{
//			feLog("DrawHydra::drawPolys bypass %d\n",vertices);
//		}

		DrawOpenGL::drawPolys(grain,vertex,normal,texture,vertices,strip,
				multicolor,color,hullPointMap,hullFacePoint,spDrawBuffer);
		return;
	}

	sp<Buffer> spBuffer=spDrawBuffer;
	if(spBuffer.isNull())
	{
		feLog("DrawHydra::drawPolys incompatible buffer\n");
		return;
	}

	if(spBuffer->m_multicolor!=multicolor || spBuffer->m_firstColor!=color[0])
	{
		spBuffer->releaseAll();
	}

	HydraDriver*& rpDriver=spBuffer->m_pDriver;
	updateDriver(rpDriver);

	GLuint& rVao=spBuffer->m_vao;
	if(!rVao)
	{
		const BWORD uvSpace=isUvSpace();
		if(uvSpace && !texture)
		{
			feLog("DrawHydra::drawPolys"
					" vertices %d in UV space without UV coordinates\n",
					vertices);
			return;
		}

		//* NOTE no refinement without hull
		const I32 refinement=hasHull? drawMode()->refinement(): 0;
		const BWORD refined=(refinement>0);

		HydraDelegate& rDelegate=rpDriver->GetDelegate();
		rDelegate.SetRefineLevel(refinement);

		spBuffer->m_multicolor=multicolor;
		spBuffer->m_firstColor=color[0];

//		const GfVec3f center=rDelegate.PopulateBasicTestSet();

		// VERTEX, UNIFORM, CONSTANT, FACEVARYING, VARYING
		const HydraDelegate::Interpolation interpolation=multicolor?
				HydraDelegate::VERTEX:
				HydraDelegate::CONSTANT;

		I32 pointCount=0;
		I32 faceCount=0;
		GfVec3f* points=NULL;
		int* numVerts=NULL;
		int* verts=NULL;
		int vertCount=0;
		GfVec3f* normals=NULL;
		BWORD allTriangles=TRUE;
		VtValue vtColor;
		VtVec4fArray vtColorArray;

		TfToken orientation = HdTokens->leftHanded;
		I32 pointIndex0= -1;
		I32 pointIndex1= -1;
		I32 pointIndex2= -1;

		if(hasHull)
		{
			const Array<I32>& rHullPointMap= *hullPointMap;
			const Array<Vector4i>& rHullFacePoint= *hullFacePoint;

			pointCount=rHullPointMap.size();
			faceCount=rHullFacePoint.size();

#if FE_DHY_DEBUG
			feLog("DrawHydra::drawPolys hull points %d faces %d\n",
					pointCount,faceCount);
#endif

			points=new GfVec3f[pointCount];
			numVerts=new int[faceCount];
			verts=new int[faceCount*4];

			if(uvSpace)
			{
				for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
				{
					const Vector2f& rUV=texture[rHullPointMap[pointIndex]];
					points[pointIndex].Set(rUV[0],rUV[1],Real(0));
				}
			}
			else
			{
				for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
				{
					points[pointIndex].Set(
							vertex[rHullPointMap[pointIndex]].temp());
				}
			}

			//* NOTE refinement generates normals (Hydra will complain)
			if(normal)
			{
				if(faceCount)
				{
					const Vector4i& rFacePoint=rHullFacePoint[0];
					pointIndex0=rHullPointMap[rFacePoint[0]];
					pointIndex1=rHullPointMap[rFacePoint[1]];
					pointIndex2=rHullPointMap[rFacePoint[2]];
				}

				if(!refined)
				{
					normals=new GfVec3f[pointCount];

					for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
					{
						normals[pointIndex].Set(
								normal[rHullPointMap[pointIndex]].temp());
					}
				}
			}

			for(I32 faceIndex=0;faceIndex<faceCount;faceIndex++)
			{
				const Vector4i& rFacePoint=rHullFacePoint[faceIndex];
				if(rFacePoint[3]<0)
				{
					numVerts[faceIndex]=3;
					verts[vertCount++]=rFacePoint[0];
					verts[vertCount++]=rFacePoint[1];
					verts[vertCount++]=rFacePoint[2];
				}
				else
				{
					numVerts[faceIndex]=4;
					verts[vertCount++]=rFacePoint[0];
					verts[vertCount++]=rFacePoint[1];
					verts[vertCount++]=rFacePoint[2];
					verts[vertCount++]=rFacePoint[3];
					allTriangles=FALSE;
				}
			}

			const BWORD boostAlpha=(drawStyle()==DrawMode::e_wireframe);

			if(multicolor)
			{
				vtColorArray.resize(pointCount);
				for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
				{
					const Color& rColor=color[rHullPointMap[pointIndex]];
					Real alpha=rColor[3];
					if(boostAlpha)
					{
						alpha=pow(alpha,0.5);
					}

					vtColorArray[pointIndex]=GfVec4f(
							rColor[0],rColor[1],rColor[2],
							fe::maximum(alpha,m_alphaMin));
				}
				vtColor = VtValue(vtColorArray);
			}
			else
			{
				const Color& rColor=color[0];
				Real alpha=rColor[3];
				if(boostAlpha)
				{
					alpha=pow(alpha,0.5);
				}

				vtColor=GfVec4f(rColor[0],rColor[1],rColor[2],
						fe::maximum(alpha,m_alphaMin));
			}
		}
		else
		{
			pointCount=vertices;
			faceCount=vertices/grain;

#if FE_DHY_DEBUG
			feLog("DrawHydra::drawPolys discrete points %d faces %d\n",
					pointCount,faceCount);
#endif

			points=new GfVec3f[pointCount];
			numVerts=new int[faceCount];
			verts=new int[pointCount];

			if(uvSpace)
			{
				for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
				{
					const Vector2f& rUV=texture[pointIndex];
					points[pointIndex].Set(rUV[0],rUV[1],Real(0));
					verts[pointIndex]=pointIndex;
				}
			}
			else
			{
				for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
				{
					points[pointIndex].Set(vertex[pointIndex].temp());
					verts[pointIndex]=pointIndex;
				}
			}

			if(normal)
			{
				normals=new GfVec3f[pointCount];

				for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
				{
					normals[pointIndex].Set(normal[pointIndex].temp());
				}

				pointIndex0=0;
				pointIndex1=1;
				pointIndex2=2;
			}

			vertCount=pointCount;
			allTriangles=(grain==3);

			for(I32 faceIndex=0;faceIndex<faceCount;faceIndex++)
			{
				numVerts[faceIndex]=grain;
			}

			if(multicolor)
			{
				vtColorArray.resize(pointCount);
				for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
				{
					const Color& rColor=color[pointIndex];

					vtColorArray[pointIndex]=GfVec4f(
							rColor[0],rColor[1],rColor[2],
							fe::maximum(Real(rColor[3]),m_alphaMin));
				}
				vtColor = VtValue(vtColorArray);
			}
			else
			{
				const Color& rColor=color[0];

				vtColor=GfVec4f(rColor[0],rColor[1],rColor[2],
						fe::maximum(Real(rColor[3]),m_alphaMin));
			}
		}

		if(pointIndex0>=0)
		{
			const SpatialVector& p0=vertex[pointIndex0];
			const SpatialVector& p1=vertex[pointIndex1];
			const SpatialVector& p2=vertex[pointIndex2];
			const SpatialVector normalCalc=
					unitSafe(cross(p1-p0,p2-p0));

			const SpatialVector& n0=normal[pointIndex0];
			const SpatialVector& n1=normal[pointIndex1];
			const SpatialVector& n2=normal[pointIndex2];
			const SpatialVector normalAve=
					unitSafe(n0+n1+n2);

			const Real normalDot=dot(normalCalc,normalAve);
			if(normalDot>Real(0))
			{
				orientation = HdTokens->rightHanded;
			}
		}

		VtVec3fArray normalArray;
		if(normals)
		{
			normalArray=_BuildArray(normals,pointCount);

			delete[] normals;
			normals=NULL;
		}

		const SdfPath id("/DrawHydra_mesh");
		const GfMatrix4f transform(1);
		const bool guide = false;
		const PxOsdSubdivTags subdivTags;
		const TfToken scheme = allTriangles?
				PxOsdOpenSubdivTokens->loop:
				PxOsdOpenSubdivTokens->catmark;
		const bool doubleSided=false;

		rDelegate.Remove(id);

		SdfPath instancerId;
		rDelegate.AddMesh(id,transform,
				_BuildArray(points, pointCount),
				_BuildArray(numVerts, faceCount),
				_BuildArray(verts, vertCount),
				normalArray,
				subdivTags,vtColor,interpolation,
				guide,instancerId,scheme,orientation,doubleSided);

		//* NOTE choices: hull, smoothHull, wire, wireOnSurf,
		//*		refined, refinedWire, refinedWireOnSurf
		rDelegate.SetReprName(id,
				drawStyle()==DrawMode::e_wireframe?
				(refined? HdTokens->refinedWire: HdTokens->wire):
				(drawStyle()==DrawMode::e_edgedSolid?
				(refined? HdTokens->refinedWireOnSurf: HdTokens->wireOnSurf):
				(refined? HdTokens->refined: HdTokens->smoothHull)));

		//* NOTE newly generated VAO acquires current state when first bound
		glGenVertexArrays(1, &rVao);
		glBindVertexArray(rVao);
		glBindVertexArray(0);

#if FE_DHY_DEBUG
		feLog("DrawHydra::drawPolys new VAO %d vertices %d\n",rVao,vertices);
#endif

		delete[] verts;
		verts=NULL;

		delete[] numVerts;
		numVerts=NULL;

		delete[] points;
		points=NULL;
	}

	updateLighting(rpDriver);

	glBindVertexArray(rVao);

//	feLog("DrawHydra::drawPolys \"%s\"\n",c_print(drawStyle()));

	const BWORD pushLineWidth=(lineWidth()!=1.0f);
//	const Real alpha=color? color[0][3]: 1.0f;
	const BWORD lit=isLit();

	setup(drawStyle(),lit,twoSidedLighting(),FALSE,
			frontfaceCulling(),backfaceCulling(),zBuffering());

//	const BWORD blending=(multicolor || alpha!=1.0f ||
//			drawStyle()==DrawMode::e_ghost);
	const BWORD blending=TRUE;

	if(blending && drawStyle()!=DrawMode::e_foreshadow)
	{
		pushBlending();
	}

	BWORD pushedLineWidth=FALSE;
	if(drawStyle()==DrawMode::e_wireframe && pushLineWidth)
	{
		glPushAttrib(GL_LINE_BIT);
		glLineWidth(lineWidth());
		pushedLineWidth=TRUE;
	}

	rpDriver->Draw();

	if(drawStyle()==DrawMode::e_foreshadow)
	{
		if(pushLineWidth)
		{
			glPushAttrib(GL_LINE_BIT);
			glLineWidth(lineWidth());
			pushedLineWidth=TRUE;
		}

		if(blending)
		{
			pushBlending();
		}

		setup(drawStyle(),lit,twoSidedLighting(),FALSE,
				frontfaceCulling(),backfaceCulling(),FALSE);

		rpDriver->Draw();

		setup(drawStyle(),lit,twoSidedLighting(),FALSE,
				frontfaceCulling(),backfaceCulling(),zBuffering());
		glDisable(GL_POLYGON_OFFSET_LINE);
	}

	if(pushedLineWidth)
	{
		//* restore line width
		glPopAttrib();
	}

	if(blending)
	{
		glPopAttrib();
	}

	glBindVertexArray(0);
}

} // namespace
