/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "viewer.pmh"

#define FE_OV_DISPLAY_FRAMERATE		TRUE

namespace fe
{

#define FRAMECOUNT_PERIOD	500.0f	//* ms

ObjectViewer::ObjectViewer(void)
{
	m_frameMS=0.0f;
	m_frameRate=0.0f;
	m_frameCount=0;
	m_frameTotal=0.0f;
}

ObjectViewer::~ObjectViewer(void)
{
#ifdef FE_CHAINSIGNALER_PROFILING
	Peeker peeker;
	peek(peeker);

	feLog("~ObjectViewer\n%s\n",peeker.str().c_str());
#endif
}

void ObjectViewer::initialize(void)
{
	m_spControllerI=registry()->create("ControllerI");
	if(!m_spControllerI.isValid())
		feX("ObjectViewer::initialize","needed to create an ControllerI");

	SpatialTransform identity;
	setIdentity(identity);

	m_spCameraOverlay=new CameraDisplay();
	m_spCameraOverlay->setName("ObjectViewer.CameraCommon overlay");
	m_spCameraOverlay->setProjection(CameraI::e_ortho);
	m_spCameraOverlay->setCameraMatrix(identity);
	m_spCameraOverlay->setRasterSpace(TRUE);
}

void ObjectViewer::bind(sp<WindowI> spWindowI)
{
	m_spDrawI->view()->setWindow(spWindowI);

	//* TODO should use signaler/scope from WindowI, not our signaler bind
//	sp<EventContextI> spEventContextI=spWindowI->getEventContextI();
}

void ObjectViewer::bind(sp<SignalerI> spSignalerI,sp<Layout> spLayout)
{
	FEASSERT(spSignalerI.isValid());

	sp<Scope> spScope=spLayout->scope();
	FEASSERT(spScope.isValid());

	if(m_spScope!=spScope)
	{
		if(m_spScope.isValid())
			feLog("%p ObjectViewer::bind() changed scope\n",this);

		m_spScope=spScope;

		sp<Layout> spEventLayout=spScope->lookupLayout(FE_EVENT_LAYOUT);
		FEASSERT(spEventLayout.isValid());

		sp<HandlerI> spHandlerI=m_spControllerI;
		FEASSERT(spHandlerI.isValid());

		spSignalerI->insert(spHandlerI,spEventLayout);
	}

	m_asViewer.bind(spScope);
}

void ObjectViewer::handle(Record &r_sig)
{
	const Real nearplane=0.1f;
	const Real farplane=1e4;

	sp<ViewI> spViewI=m_spDrawI->view();
	sp<CameraDisplay> spCameraDisplay=spViewI->cameraDisplay();
	sp<WindowI> spWindowI=spViewI->window();
	Box2i viewport=spWindowI->geometry();
	viewport[0]=0;
	viewport[1]=0;

	I32 layer= -1;
	if(m_asViewer.viewer_layer.check(r_sig))
	{
		layer=m_asViewer.viewer_layer(r_sig);
	}

	if(layer==0)
	{
		const Real ms=m_ticker.timeMS();
		m_frameTotal+=ms-m_frameMS;
		m_frameCount++;
		if(m_frameTotal>FRAMECOUNT_PERIOD)
		{
			m_frameRate=m_frameCount*1e3f/m_frameTotal;
			m_frameTotal=0.0f;
			m_frameCount=0;
		}
		m_frameMS=ms;
	}

	const BWORD spatial=(m_asViewer.viewer_spatial.check(r_sig) &&
			m_asViewer.viewer_spatial(r_sig));

	if(spatial)
	{
		sp<CameraI> spCamera=m_spControllerI->camera();

		const CameraI::Projection projection=spCamera->projection();
		const Vector2 fov=spCamera->fov();
		const SpatialTransform& cameraMatrix=spCamera->cameraMatrix();
		Real zoom=1.0;
		Vector2 center(0.0,0.0,0.0);

		sp<CameraCommon> spCameraCommon=spCamera;
		if(spCameraCommon.isValid())
		{
			spCameraCommon->getOrtho(zoom,center);
		}

		Real fovy=fov[1];
		if(fov[0]/fov[1] > width(viewport)/Real(height(viewport)))
		{
			fovy=fov[0]*height(viewport)/Real(width(viewport));
		}

		spCameraDisplay->setViewport(viewport);
		spCameraDisplay->setCameraMatrix(cameraMatrix);
		spCameraDisplay->setPerspective(fovy,nearplane,farplane);
		spCameraDisplay->setOrtho(zoom,center);
		spViewI->use(projection);

		signal(r_sig);

		//* NOTE store perspective info for ViewI::project()
		m_spCameraOverlay->setCameraMatrix(cameraMatrix);
		m_spCameraOverlay->setPerspective(fovy,nearplane,farplane);
	}
	else
	{
		m_spCameraOverlay->setViewport(viewport);
		spViewI->setCameraDisplay(m_spCameraOverlay);

#if FE_OV_DISPLAY_FRAMERATE
		if(layer>0)
		{
			const Color pink(1.0f,0.5f,0.5f);
			const SpatialVector corner(10.0f,10.0f,0.0f);

			String text;
			text.sPrintf("%.0f FPS",frameRate());
			m_spDrawI->drawAlignedText(corner,text,pink);
		}
#endif

		signal(r_sig);

		spViewI->setCameraDisplay(spCameraDisplay);
	}
}

} /* namespace */
